/* bigint.c   by Michael Thorpe   2017-05-29 */

#include <ctype.h>
#include <gmp.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include "lisp.h"

static obj *gmperror;

/* Takes an integer and upconverts it to a bigint */
static obj *promote(obj *in) {
	obj *out;

	out=new_object(&bigint_objtype);
	if(out) {
/* FIXME: What happens if GMP runs out of memory here? */
		mpz_init_set_si(out->value.bi,in->value.i);
	}
	DECREF(in);
	return(out);
}

/* Takes a newly-minted bigint and downconverts it to an integer if possible */
static obj *reduce(obj *in) {
	signed long i;

	if(!mpz_fits_slong_p(in->value.bi))
		return(in);
	i=mpz_get_si(in->value.bi);
	mpz_clear(in->value.bi);
	in->type=&integer_objtype;
	in->value.i=i;
	return(in);
}

obj *addb(obj *a,obj *b) {
	obj *r;

	if(&integer_objtype==a->type) {
		a=promote(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
	if(a->type != &bigint_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(&integer_objtype==b->type) {
		b=promote(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
	if(b->type != &bigint_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
	r=new_object(&bigint_objtype);
	if(!r) {
		DECREF(a);
		DECREF(b);
		return(0);
	}
	mpz_init(r->value.bi);
/* FIXME: What happens if GMP runs out of memory here? */
	mpz_add(r->value.bi,a->value.bi,b->value.bi);
	DECREF(a);
	DECREF(b);
	return(reduce(r));
}

obj *bitselectb(obj *a,obj *if0,obj *if1) {
	obj *r;

	if(&integer_objtype==a->type) {
		a=promote(a);
		if(!a) {
			DECREF(if0);
			DECREF(if1);
			return(0);
		}
	}
	if(a->type != &bigint_objtype) {
		DECREF(if0);
		DECREF(if1);
		return(throwtypeerror(a));
	}
	if(&integer_objtype==if0->type) {
		if0=promote(if0);
		if(!if0) {
			DECREF(a);
			DECREF(if1);
			return(0);
		}
	}
	if(if0->type != &bigint_objtype) {
		DECREF(a);
		DECREF(if1);
		return(throwtypeerror(if0));
	}
	if(&integer_objtype==if1->type) {
		if1=promote(if1);
		if(!if1) {
			DECREF(a);
			DECREF(if0);
			return(0);
		}
	}
	if(if1->type != &bigint_objtype) {
		DECREF(a);
		DECREF(if0);
		return(throwtypeerror(if1));
	}
/*
 * For each bit in a, selects the corresponding bit in if0 if it's 0 or
 * the corresponding bit in if1 if it's 1.  For instance:
 *   (bitwise-not a) is (bitselect a -1 0)
 *   (bitwise-and a b) is (bitselect a 0 b)
 *   (bitwise-or a b) is (bitselect a -1 b)
 *   (bitwise-xor a b) is (bitselect a b (bitwise-not b))
 * We use the equivalence that (if0&!a)|(if1&a) == (if0^if1)&a)^if0.
*/
	r=new_object(&bigint_objtype);
	if(r) {
		mpz_init(r->value.bi);
		mpz_xor(r->value.bi,if0->value.bi,if1->value.bi);
		mpz_and(r->value.bi,r->value.bi,a->value.bi);
		mpz_xor(r->value.bi,r->value.bi,if0->value.bi);
		r=reduce(r);
	}
	DECREF(a);
	DECREF(if0);
	DECREF(if1);
	return(r);
}

obj *div2b(obj *a,obj *b) {
	obj *r;

	if(&integer_objtype==b->type) {
		if(0==b->value.i) {
			DECREF(b);
			return(throw(INCREF(divideby0error),a));
		}
		b=promote(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
	if(b->type != &bigint_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
	if(&integer_objtype==a->type) {
		a=promote(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
	if(a->type != &bigint_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	r=new_object(&bigint_objtype);
	if(!r) {
		DECREF(a);
		DECREF(b);
		return(0);
	}
	mpz_init(r->value.bi);
/* FIXME: What happens if GMP runs out of memory here? */
	mpz_tdiv_q(r->value.bi,a->value.bi,b->value.bi);
	DECREF(a);
	DECREF(b);
	return(reduce(r));
}

/* Returns #t, #f, or the input number if it's not sure */
obj *isprime(obj *n) {
	int res;

	if(&integer_objtype==n->type) {
		n=promote(n);
		if(!n)
			return(0);
	}
	if(n->type != &bigint_objtype)
		return(throwtypeerror(n));
	res=mpz_probab_prime_p(n->value.bi,32);
	if(2==res) {
		DECREF(n);
		return(INCREF(&trueobj));
	}
	if(0==res) {
		DECREF(n);
		return(INCREF(&nilobj));
	}
/* We almost always don't need to reduce, but just in case: */
	return(reduce(n));
}

obj *mulb(obj *a,obj *b) {
	obj *r;

	if(&integer_objtype==a->type) {
		a=promote(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
	if(a->type != &bigint_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(&integer_objtype==b->type) {
		b=promote(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
	if(b->type != &bigint_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
	r=new_object(&bigint_objtype);
	if(!r) {
		DECREF(a);
		DECREF(b);
		return(0);
	}
	mpz_init(r->value.bi);
/* FIXME: What happens if GMP runs out of memory here? */
	mpz_mul(r->value.bi,a->value.bi,b->value.bi);
	DECREF(a);
	DECREF(b);
/* We only need to reduce() here when -LONG_MIN is multiplied by -1: */
	return(reduce(r));
}

obj *nextprime(obj *n) {
	obj *r;

	if(&integer_objtype==n->type) {
		n=promote(n);
		if(!n)
			return(0);
	}
	if(n->type != &bigint_objtype)
		return(throwtypeerror(n));
	r=new_object(&bigint_objtype);
	if(r) {
		mpz_init(r->value.bi);
		mpz_nextprime(r->value.bi,n->value.bi);
	}
	DECREF(n);
	return(r);
}

#define NUMCMP_LT 1
#define NUMCMP_EQ 2
#define NUMCMP_GT 4
obj *numcmpb(obj *a,obj *b,int desired) {
	int cmp;

	if(&integer_objtype==a->type) {
		a=promote(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
	if(a->type != &bigint_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(&bigint_objtype==b->type)
		cmp=mpz_cmp(a->value.bi,b->value.bi);
	else if(&integer_objtype==b->type)
		cmp=mpz_cmp_si(a->value.bi,b->value.i);
	else {
		DECREF(a);
		return(throwtypeerror(b));
	}
	DECREF(a);
	DECREF(b);
	if(cmp<0)
		cmp=NUMCMP_LT;
	else if(0==cmp)
		cmp=NUMCMP_EQ;
	else
		cmp=NUMCMP_GT;
	if(cmp&desired)
		return(INCREF(&trueobj));
	else
		return(INCREF(&nilobj));
}

obj *string_to_bigint(obj *str) {
	int i;
	obj *o;
	char *buf;

	if(str->type != &string_objtype)
		return(throwtypeerror(str));
	if(0==str->value.s.len)
		return(throwtypeerror(str));
	for(i=('-'==str->value.s.data[0]);i<str->value.s.len;i++) {
		if(!isdigit(str->value.s.data[i]))
			return(throwtypeerror(str));
	}
	buf=(char *)malloc(str->value.s.len+1);
	if(!buf) {
		DECREF(str);
		return(throwoom());
	}
	memcpy(buf,str->value.s.data,str->value.s.len);
	buf[str->value.s.len]='\0';
	DECREF(str);
	o=new_object(&bigint_objtype);
	if(!o) {
		free(buf);
		return(0);
	}
	if(mpz_init_set_str(o->value.bi,buf,10)) {
		free(buf);
		DECREF(o);
		return(INCREF(gmperror));
	}
	free(buf);
	return(reduce(o));
}

obj *subb(obj *a,obj *b) {
	obj *r;

	if(&integer_objtype==a->type) {
		a=promote(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
	if(a->type != &bigint_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(&integer_objtype==b->type) {
		b=promote(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
	if(b->type != &bigint_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
	r=new_object(&bigint_objtype);
	if(!r) {
		DECREF(a);
		DECREF(b);
		return(0);
	}
	mpz_init(r->value.bi);
/* FIXME: What happens if GMP runs out of memory here? */
	mpz_sub(r->value.bi,a->value.bi,b->value.bi);
	DECREF(a);
	DECREF(b);
	return(reduce(r));
}

#ifdef USE_FLOAT
obj *bigint_to_float(obj *bi) {
	obj *f;

	if(bi->type != &bigint_objtype)
		return(throwtypeerror(bi));
	f=new_object(&float_objtype);
	if(f)
		f->value.f=mpz_get_d(bi->value.bi);
	DECREF(bi);
	return(f);
}
#endif

static void bigint_nuke(obj *in) {
	mpz_clear(in->value.bi);
}

static obj *bigint_tostring(obj *in) {
	obj *out;
	char *s;
	int i;

	s=mpz_get_str(0,10,in->value.bi);
	if(!s)
		return(throw(gmperror,in));
	DECREF(in);
	i=strlen(s);
	out=new_string(s,i);
	free(s);
	return(out);
}

static int bigint_init() {
	if(!(gmperror=new_word("gmp-error",9))
#ifdef USE_FLOAT
	|| store_builtin_ns1("bigint->float",bigint_to_float)
#endif
	|| store_builtin_ns1("prime?",isprime)
	|| store_builtin_ns1("next-prime",nextprime))
		return(1);
	return(0);
}

static void bigint_exit() {
	DECREF(gmperror); gmperror=0;
}

objtype bigint_objtype={
	name: "bigint",
	predropobj: bigint_nuke,
	tostring: bigint_tostring,
	init: bigint_init,
	exit: bigint_exit,
};

